local Utopia = LibStub("AceAddon-3.0"):GetAddon("Utopia")
if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 38 $")

local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour
local rotate = Utopia.rotate
local band = bit.band
local savageCombat = GetSpellInfo(58413)
local bloodFrenzy = GetSpellInfo(29859)

----------------- BUG FIX ----- Remove this when WoW includes Savage Combat/Blood Frenzy in combat log --------------
local SChaxCache = setmetatable({},{
	__index = function(self,key)
		local val, class = UnitClass(key)
		if (class == "ROGUE") then
			val = Utopia:HasTalent(key, GetSpellInfo(58413))	-- Savage Combat
		elseif (class == "WARRIOR") then
			val = Utopia:HasTalent(key, GetSpellInfo(29859))	-- Blood Frenzy
		end
		self[key] = val or false
		return val
	end
})
--------------END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

local engine = CreateFrame("Frame", "UtopiaDebuffActivity")
--@debug@
E = engine
function engine:Print(...)
	ChatFrame1:AddMessage("|cFFFFFF80Engine: |r"..format(...))
end
--@no-debug@

engine:SetScript("OnEvent", function(self, event, ...)
	local f = engine[event]
	if (f) then
		f(self, ...)
	else
		error("No handler for "..event)
	end
end)

-- IsBlacklistedMobID
local IsBlacklistedMobID
do
	local blackListMobIDs = {
		[19833] = true,		-- Hunter, Snake Trap snake
		[19921] = true,		-- Hunter, Snake Trap snake
	}

	function IsBlacklistedMobID(guid)
		return blackListMobIDs[tonumber(strsub(guid, -12,-7),16)] == true
	end	
end

-- Start
function engine:Start()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("UNIT_FLAGS")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_NAME_UPDATE")
	self:RegisterEvent("UNIT_PET")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")

	self.mobs = new()
	self.players = new()
	self:ScanRoster()
	self:ResetInCombatForAll()
end

-- Stop
function engine:Stop()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("UNIT_FLAGS")
	self:UnregisterEvent("RAID_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:UnregisterEvent("UNIT_NAME_UPDATE")
	self:UnregisterEvent("UNIT_PET")
	self:UnregisterEvent("UNIT_AURA")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("PLAYER_FOCUS_CHANGED")

	self.mobs = deepDel(self.mobs)
	self.players = deepDel(self.players)
	self.combat = deepDel(self.combat)
end

local function checkRosterUpdate(self, elapsed)
	if (self.rosterUpdate) then
		self.rosterUpdate = nil
		self:SetScript("OnUpdate", nil)
		self:ScanRoster()
	end
end

function engine:RAID_ROSTER_UPDATE()
	self.rosterUpdate = true
	self:SetScript("OnUpdate", checkRosterUpdate)
end
engine.PARTY_MEMBERS_CHANGED = engine.RAID_ROSTER_UPDATE

-- PLAYER_TARGET_CHANGED
function engine:PLAYER_TARGET_CHANGED()
	self.targetGUID = UnitGUID("target")
	self:CheckTargetDebuffs("target")
end

-- PLAYER_FOCUS_CHANGED
function engine:PLAYER_FOCUS_CHANGED()
	self.focusGUID = UnitGUID("focus")
	self:CheckTargetDebuffs("focus")
end

-- CheckTargetDebuffs
function engine:CheckTargetDebuffs(target)
	-- Validate the target and load up debuffs if it has any
	-- Mark it as temporary and dump them again if we drop target out of combat

	for guid,mob in pairs(self.mobs) do
		if (mob.temp) then
			for debuffName,debuff in pairs(mob.debuffs) do
				if (debuff.temp) then
					--self:Print("Cleaning temp debuff |cFFFF8080%s|r from |cFFFF0000%s", debuffName, mob.name)
					mob.debuffs[debuffName] = del(debuff)
				end
			end
			mob.temp = nil
			if (not next(mob.debuffs)) then
				--self:Print("Removing |cFFFF0000%s|r because of empty .debuffs", mob.name)
				self.mobs[guid] = del(self.mobs[guid])
			end
		end
	end
	--self:ResetInCombatForAll()
	--if (not next(self.combat)) then
	--	self:Cleanup()
	--end

	if (not UnitExists(target) or UnitIsDeadOrGhost(target) or not UnitCanAttack("player", target)) then
		return
	end

	local lookup = Utopia.lookup
	if (not lookup) then
		return
	end

	local guid = UnitGUID(target)
	local mob = self.mobs[guid]
	local debuffs = mob and mob.debuffs

	for i = 1,100 do
	local name, rank, tex, count, debuffType, duration, endTime, isMine, isStealable = UnitDebuff(target, i)
		if (name == nil) then
			break
		end

		local def = lookup[name]
		if (def) then
			if (not mob) then
				--self:Print("Creating temporary mob for |cFFFF0000%s", UnitName(target))
				mob = new()
				self.mobs[guid] = mob
				mob.name = UnitName(target)
				mob.guid = guid
				mob.debuffs = new()
				debuffs = mob.debuffs
			end

			local debuff = debuffs[name]
			if (not debuff) then
				-- Debuff exists on target that we've ignored in combat log

				--self:Print("- %s: New TEMP debuff: %s %s", UnitName(target), name, rank or "NoRank")
				debuff = new()
				debuffs[name] = debuff

				debuff.name = name
				debuff.rank = rank and tonumber(strmatch(rank, "(%d)"))
				debuff.source = isMine and UnitGUID("player") or nil
				debuff.temp = true
				mob.temp = true
				debuff.stacks = count
			end
		end
	end
end

-- UNIT_NAME_UPDATE
function engine:UNIT_NAME_UPDATE(unit)
	local guid = UnitGUID(unit)
	local player = self.players[guid]
	if (player) then
		player.name = UnitFullName(unit)
	else
		local mob = self.mobs[guid]
		if (mob) then
			mob.name = UnitName(unit)
		end
	end
end

-- UNIT_PET
function engine:UNIT_PET(unit)
	--if (not UnitInParty(unit) and not UnitInRaid(unit)) then
	--	return
	--end

	local ownerGUID = UnitGUID(unit)
	if (not ownerGUID) then
		-- Probably zoning
		return 
	end

	local player = self.players[ownerGUID]
	if (not player) then
		-- We don't care about people outside the raid
		return
	end

	local petid
	if (unit == "player") then
		petid = "pet"
	else
		petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
	end

	if (UnitExists(petid) and UnitCreatureFamily(petid)) then
		local petGUID = UnitGUID(petid)
		if (not self.players[petGUID]) then
			self:AddPlayer(petid, unit)
			player.pet = petGUID
			--self:Print("Added pet for %s called %s", UnitName(unit), UnitName(petid))
		end
	else
		local petGUID = player.pet
		if (petGUID) then
			player.pet = nil
			if (self.players[petGUID]) then
				--self:Print("Removed pet from %s", UnitName(unit))
				self.players[petGUID] = del(self.players[petGUID])
			end
		end
	end
end

-- AddPlayer
function engine:AddPlayer(unit, ownerid)
	local name = UnitFullName(unit)
	local guid = UnitGUID(unit)

	local n = self.oldRoster and self.oldRoster[guid]
	if (n) then
		self.oldRoster[guid] = nil
	else
		n = new()
	end
	self.players[guid] = n

	n.name = name
	n.guid = guid
	n.ownerid = ownerid

	return n
end

-- ScanRoster
function engine:ScanRoster()
	if (not self.players) then
		return
	end

	self.oldRoster = self.players
	self.players = new()
	self.scRogues = 0		-- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	self.bfWarriors = 0

	for unit in Utopia:IterateRoster() do
		local player = self:AddPlayer(unit)

		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		local _, class = UnitClass(unit)
		if (class == "ROGUE") then
			if (Utopia.talents and Utopia.talents[UnitFullName(unit)]) then
				if (Utopia:HasTalent(unit, GetSpellInfo(58413))) then
					self.scRogues = self.scRogues + 1
				end
			--else
			--	self.scRogues = self.scRogues + 1
			end
		elseif (class == "WARRIOR") then
			-- Dittoe for Warrior's Blood Frenzy
			if (Utopia.talents and Utopia.talents[UnitFullName(unit)]) then
				if (Utopia:HasTalent(unit, GetSpellInfo(29859))) then
					self.bfWarriors = self.bfWarriors + 1
				end
			--else
			--	self.scRogues = self.scRogues + 1
			end
		end
		-------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log -------------

		local petid
		if (unit == "player") then
			petid = "pet"
		else
			petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
		end

		if (UnitExists(petid) and UnitCreatureFamily(petid)) then
			self:AddPlayer(petid, unit)
			player.pet = UnitGUID(petid)
		end
	end

	----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	if (self.scRogues > 0 or self.bfWarriors > 0) then
		self:RegisterEvent("UNIT_AURA")
	else
		self:UnregisterEvent("UNIT_AURA")
	end
	-------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log -------------

	self.oldRoster = deepDel(self.oldRoster)
end

-- COMBAT_LOG_EVENT_UNFILTERED
local mask = COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID + COMBATLOG_OBJECT_AFFILIATION_MINE
local dstMask = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL + COMBATLOG_OBJECT_CONTROL_NPC

function engine:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, ...)
	--for i = 1,select("#", ...) do
	--	local str = tostring(select(i, ...))
	--	if (str == "Savage Combat") then
	--		self:Print("|cFFFF0000%s in event %s", str, event)
	--		X = {timestamp, event, ...}
	--	end
	--end

	local ev = self.clEvents[event]
	if (ev) then
		ev(self, timestamp, event, ...)
	end
end

engine.clEvents = {}
-- SPELL_AURA_APPLIED
function engine.clEvents:SPELL_AURA_APPLIED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (band(srcFlags, mask) ~= 0) then
		if (dstGUID ~= self.targetGUID and dstGUID ~= self.focusGUID) then
			if (band(dstFlags, dstMask) == 0) then
				return
			end
		end

		local player = self.players[srcGUID]
		if (not player) then
			-- Probably a hunter snake trap pet or something else minor
			return
		end

		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		if (self.scRogues > 0 and spellSchool == SCHOOL_MASK_NATURE) then
			-- When a rogue with Savage Combat talent applies a poison, we remember who
			-- they are and the target they hit for next UNIT_AURA check... YUK!
			local _, class = UnitClass(srcName)
			if (class == "ROGUE") then
				if (SChaxCache[srcName]) then				-- Talent check cached for speed
					if (not self[savageCombat]) then
						self[savageCombat] = new()
					end
					--self:Print("Flagging |cFFFF8080Savage Combat|r potential for |cFFFFFF80%s|r on |cFFFF0000%s", srcName, dstName)
					self[savageCombat][dstGUID] = srcGUID
					return
				end
			end
		elseif (self.bfWarriors > 0 and spellSchool == SCHOOL_MASK_PHYSICAL) then
			if (spellName == GetSpellInfo(772) or spellId == 12721) then			-- Rend & Deep Wounds
				-- Same for Warriors who apply Rend
				local _, class = UnitClass(srcName)
				if (class == "WARRIOR") then
					if (SChaxCache[srcName]) then				-- Talent check cached for speed
						if (not self[bloodFrenzy]) then
							self[bloodFrenzy] = new()
						end
						--self:Print("Flagging |cFFFF8080Blood Frenzy|r potential for |cFFFFFF80%s|r on |cFFFF0000%s", srcName, dstName)
						self[bloodFrenzy][dstGUID] = srcGUID
						return
					end
				end
			end
		end
		------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

		------- And remove this if they put Judgements of the Just and Master Poisoner in too.... ------
		local mapping = Utopia.mappingLookup[spellName]
		local mappedSpellName
		if (mapping) then
			--self:Print("Possible mapping for |cFFFF0000%s|r on |cFFFF8080%s|r from |cFFFFFF80%s", spellName, dstName, srcName)
			local _, class = UnitClass(srcName)
			if (mapping.class == class) then
				--self:Print("- Class matches")
				if (Utopia:HasTalent(srcName, mapping.requiredTalent)) then
					--self:Print("- Source has the |cFF80FF80%s|r talent", mapping.requiredTalent)
					mappedSpellName = spellName
					spellName = mapping.key
				--else
				--	self:Print("- Source does not have the |cFF80FF80%s|r talent", mapping.requiredTalent)
				end
			--else
			--	self:Print("- class does not match")			-- Won't happen
			end
		end

		if (not Utopia.lookup[spellName]) then
			return
		end

		--self:Print("%s: |cFFFFFF80%s|r applies |cFFFF8080%s|r to |cFFFF0000%s|r - School = %s", event, srcName, spellName, dstName, tostring(spellSchool))
		local mob = self.mobs[dstGUID]
		if (not mob) then
			--self:Print("- New Mob: |cFFFF0000%s", dstName)
			mob = new()
			self.mobs[dstGUID] = mob

			mob.name = dstName
			mob.guid = dstGUID
			mob.debuffs = new()
		end

		if (mappedSpellName) then
			if (not self.mappings) then
				self.mappings = new()
			end
			if (not self.mappings[dstGUID]) then
				self.mappings[dstGUID] = new()
			end
			if (not self.mappings[dstGUID][srcGUID]) then
				self.mappings[dstGUID][srcGUID] = new()
			end
			--self:Print("- Added mapping for |cFFFF8080%s|r to |cFFFF8080%s", mappedSpellName, spellName)
			self.mappings[dstGUID][srcGUID][mappedSpellName] = spellName
		end

		local debuff = mob.debuffs[spellName]
		if (not debuff) then
			--self:Print("- New debuff for |cFFFF0000%s|r: |cFFFF8080%s", dstName, spellName)
			debuff = new()
			mob.debuffs[spellName] = debuff

			debuff.name = spellName
			debuff.id = spellId
			debuff.mapped = mappedSpellName
		end

		debuff.source = srcGUID
		debuff.stacks = 1
		debuff.startTime = GetTime()

		Utopia:OnDebuffApplied(dstGUID, spellName, debuff)
	end
end

-- SPELL_AURA_APPLIED_DOSE
function engine.clEvents:SPELL_AURA_APPLIED_DOSE(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (not Utopia.lookup[spellName]) then
		return
	end
	--self:Print("%s: |cFFFFFF80%s|r applies |cFFFF8080%s|r to |cFFFF0000%s|r", event, srcName, spellName, dstName)
	if (band(dstFlags, dstMask) == 0) then
		return
	end

	local mob = self.mobs[dstGUID]
	if (not mob) then
		-- self:Print("SPELL_AURA_APPLIED_DOSE(1) |cFFFF8080%s|r without a SPELL_AURA_APPLIED for %s", spellName or "nil", dstName or "nil")
		-- Probably targetted a mob being killed by someone outside group
		return
	end

	local debuff = mob.debuffs[spellName]
	if (debuff) then
		debuff.startTime = GetTime()
		debuff.stacks = (debuff.stacks or 0) + 1
		--self:Print("- Stacks for |cFFFF0000%s's|r |cFFFF8080%s|r incremented to |cFF00FF00%d|r", dstName, spellName, debuff.stacks)

		Utopia:OnDebuffApplied(dstGUID, spellName, debuff, true)
--	else
--		Probably helping on a mob being killed by another group
	end
end


-- SPELL_AURA_REMOVED
function engine.clEvents:SPELL_AURA_REMOVED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (band(srcFlags, mask) ~= 0) then
		if (not Utopia.lookup[spellName]) then
			if (not Utopia.mappingLookup[spellName]) then
				return
			end
		end
		if (dstGUID ~= self.targetGUID and dstGUID ~= self.focusGUID) then
			if (band(dstFlags, dstMask) == 0) then
				return
			end
		end

		if (srcGUID) then
			local destMapping = self.mappings and self.mappings[dstGUID]
			if (destMapping) then
				--self:Print("Checking mappings for |cFFFF0000%s", dstName)
				local srcMapping = destMapping[srcGUID]
				if (srcMapping) then
					--self:Print("- Found mappings from |cFFFFFF80%s", srcName)
					local spell = srcMapping[spellName]
					if (spell) then
						--self:Print("Mapped spell removed from |cFFFF0000%s|r: |cFFFF8080%s|r -> |cFFFF8080%s|r", dstName, spellName, spell)
						srcMapping[spellName] = nil

						if (not next(srcMapping)) then
							-- No more mappings from this class on this target, so we can remove the
							-- pseudo buff. This makes the assumption that there is only one mapping
							-- spell per class, which is safe given the dataset at this time
							--self:Print("- Removing mapping for |cFFFF8080%s|r to |cFFFF8080%s", spellName, spell)
							spellName = spell

							destMapping[srcGUID] = del(srcMapping)
							if (not next(destMapping)) then
								self.mappings[dstGUID] = del(destMapping)
								if (not next(self.mappings)) then
									self.mappings = del(self.mappings)
								end
							end
						--else
						--	self:Print("- Keeping mapping for |cFFFF8080%s|r", spell)
						end
					end
				end
			end
		end

		--self:Print("%s: |cFFFFFF80%s's|r debuff of |cFFFF8080%s|r removed from |cFFFF0000%s|r", event, srcName, spellName, dstName)
		local player = self.players[srcGUID]
		if (player) then
			local mob = self.mobs[dstGUID]
			if (mob) then
				local debuff = mob.debuffs[spellName]
				if (debuff) then
					Utopia:OnDebuffRemoved(dstGUID, spellName, debuff)
					mob.debuffs[spellName] = nil
				--else
					-- Probably ran out of combat and the mob's debuffs auto expired. This is ok.
				end
			--else
				-- Probably ran out of combat and the mob's debuffs auto expired. This is ok.
			end
		--else
			-- Probably debuff from a hunter snake trap pet expiring
		end
	end
end

-- SPELL_AURA_REMOVED_DOSE
function engine.clEvents:SPELL_AURA_REMOVED_DOSE(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (not Utopia.lookup[spellName]) then
		return
	end
	if (band(dstFlags, dstMask) == 0) then
		return
	end

	local mob = self.mobs[dstGUID]
	if (mob) then
		local debuff = mob.debuffs[spellName]
		if (debuff) then
			debuff.stacks = debuff.stacks - 1
			Utopia:OnDebuffRemoved(dstGUID, spellName, debuff, true)
--@debug@
			if (debuff.stacks < 1) then
				self:Print("|cFFFF8080%s|r Stacks decremented to |cFFFF0000%d|r on |cFFFF0000%s", spellName, debuff.stacks, dstName)
			end
		else
			self:Print("SPELL_AURA_REMOVED_DOSE: with no engine.mobs['%s']['%s']", dstName, spellName)
--@no-debug@
		end
--@debug@
	else
		self:Print("SPELL_AURA_REMOVED_DOSE: with no engine.mobs['%s']", dstName)
--@no-debug@
	end
end

----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
function engine:ApplyHackedAura(unit, spellName, onoff, ...)
	local dstGUID = UnitGUID(unit)
	local srcGUID = self[spellName] and self[spellName][dstGUID]

	if (onoff) then
		if (not self.has) then
			self.has = new()
		end
		if (not self.has[spellName]) then
			self.has[spellName] = new()
		end

		if (not self.has[spellName][dstGUID]) then
			self.has[spellName][dstGUID] = srcGUID

			local player = self.players[srcGUID]
			if (player) then
				local mob = self.mobs[dstGUID]
				if (not mob) then
					mob = new()
					self.mobs[dstGUID] = mob
	
					mob.name = UnitName(unit)
					mob.guid = dstGUID
					mob.debuffs = new()
				end
	
				local debuff = mob.debuffs[spellName]
				if (not debuff) then
					debuff = new()
					mob.debuffs[spellName] = debuff
				end
	
				debuff.name = spellName
				debuff.id = select(onoff, ...)
				debuff.source = srcGUID
				debuff.stacks = 1
	
				Utopia:OnDebuffApplied(dstGUID, spellName, debuff)
			end
		end
	else
		local mob = self.mobs[dstGUID]
		if (mob) then
			local debuff = mob.debuffs[spellName]
			if (debuff) then
				if (self.has and self.has[spellName] and self.has[spellName][dstGUID]) then
					self.has[spellName][dstGUID] = nil
				end

				Utopia:OnDebuffRemoved(dstGUID, spellName, debuff)
				mob.debuffs[spellName] = nil
			end
		end
	end
end

-- UNIT_AURA
function engine:UNIT_AURA(unit)
	if (not UnitCanAttack("player", unit)) then
		return
	end

	local foundSC, foundBF
	for i = 1,100 do
		local name, rank, tex = UnitDebuff(unit, i)
		if (name == nil) then
			break
		elseif (name == savageCombat) then
			foundSC = tonumber(rank:match("(%d)")) or 2
		elseif (name == bloodFrenzy) then
			foundBF = tonumber(rank:match("(%d)")) or 2
		end
	end

	if (self.scRogues > 0) then
		self:ApplyHackedAura(unit, savageCombat, foundSC, 51682, 58413)
	end
	if (self.bfWarriors > 0) then
		self:ApplyHackedAura(unit, bloodFrenzy, foundBF, 29836, 29859)
	end
end
------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

-- ResetInCombatForAll
function engine:ResetInCombatForAll()
	del(self.combat)
	self.combat = new()
	for unit in Utopia:IterateRoster() do
		self.combat[UnitGUID(unit)] = UnitAffectingCombat(unit)
	end
end

-- UNIT_FLAGS
function engine:UNIT_FLAGS(unit)
	local guid = UnitGUID(unit)
	if (UnitAffectingCombat(unit)) then
		if (UnitInParty(unit) or UnitPlayerOrPetInRaid(unit) or UnitPlayerOrPetInParty(unit)) then
			if (not self.combat[guid]) then
				self.combat[guid] = true
			end
		end
	else
		if (self.combat[guid]) then
			self.combat[guid] = nil

			if (not next(self.combat) or UnitIsUnit(unit, "player")) then
				-- Nothing flagged as in-combat, but we double check to be sure
				-- for cases where ppl become IsVisible() and we don't see a
				-- UNIT_FLAGS event for them. Also checking this when I go out
				-- of combat to be doubly sure things don't go adrift over time
				self:ResetInCombatForAll()
				if (not next(self.combat)) then
					self:Cleanup()
				end
			end
		end
	end
end

-- Cleanup
function engine:Cleanup()
	deepDel(self.mobs)
	self.mobs = new()

	----------------- BUG FIX ----- Remove this when WoW includes Savage Combat/Blood Frenzy in combat log --------------
	self[savageCombat] = del(self[savageCombat])
	self[bloodFrenzy] = del(self[bloodFrenzy])
	self.has = deepDel(self.has)
	for k in pairs(SChaxCache) do
		SChaxCache[k] = nil
	end
	------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	
	self.mappings = deepDel(self.mappings)
end

-- GetMobByUnit
function engine:GetMobByUnit(unit)
	return engine.mobs and engine.mobs[UnitGUID(unit)]
end

--@debug@
-- ShowFunc
-- Debug dump of all current debuffs on all targets
function engine:ShowFunc()
	if (self.mobs) then
		for guid, mob in pairs(self.mobs) do
			self:Print("|cFFFF0000%s|cFF808080 (%s)", mob.name, guid)

			if (next(mob.debuffs)) then
				for debuffName, debuff in pairs(mob.debuffs) do
					local playerGUID = debuff.source
					local player = engine.players[playerGUID]

					if (player) then
						self:Print("- |cFFFFFF80%s|r applied: |cFFFF8080%s|r (id=%d) (|cFF80FF80%s|r stacks)", player.name, debuffName, debuff.id, debuff.stacks)
					else
						self:Print("- |cFFFF00FFUnknown|r applied %s (id=%d) (|cFF80FF80%s|r stacks)", debuffName, debuff.id, debuff.stacks)
					end
				end
			else
				self:Print("- No debuffs")
			end
		end
	end
end

local function UE()
	engine:ShowFunc()
end

SlashCmdList["UE"] = UE
SLASH_UE1 = "/ue"
--@no-debug@
